import os

from gettext import gettext as _

from django.conf import settings
from django.core.management import BaseCommand, CommandError
from django.db.models import Q, Sum
from pulpcore import constants
from pulpcore.app import pulp_hashlib
from pulpcore.plugin.models import (
    Artifact,
    Content,
    ContentArtifact,
    RemoteArtifact,
    RepositoryVersion,
)

import logging

log = logging.getLogger("")

CHUNK_SIZE = 1024 * 1024  # 1 Mb


class Command(BaseCommand):
    """
    Django management command for populating or removing checksums on artifacts.
    """

    help = _("Handle missing and forbidden checksums on the artifacts")

    def add_arguments(self, parser):
        parser.add_argument("--report", action="store_true")
        parser.add_argument("--checksums", help=_("Comma separated list of checksums to evaluate"))

    def _print_out_repository_version_hrefs(self, repo_versions):
        """
        Print out repository version href from a query of repo versions.

        Args:
            repo_versions (django.db.models.QuerySet): repository versions query

        Returns None
        """
        for repo_version in repo_versions.iterator():
            self.stdout.write(
                _(
                    "/repositories/{plugin}/{type}/{pk}/versions/{number}/".format(
                        plugin=repo_version.repository.pulp_type.split(".")[0],
                        type=repo_version.repository.pulp_type.split(".")[1],
                        pk=str(repo_version.repository.pk),
                        number=repo_version.number,
                    )
                )
            )

    def _show_on_demand_content(self, checksums):
        query = Q(pk__in=[])
        for checksum in checksums:
            query |= Q(**{f"{checksum}__isnull": False})

        remote_artifacts = RemoteArtifact.objects.filter(query).filter(
            content_artifact__artifact__isnull=True
        )
        ras_size = remote_artifacts.aggregate(Sum("size"))["size__sum"]

        content_artifacts = ContentArtifact.objects.filter(remoteartifact__pk__in=remote_artifacts)
        content = Content.objects.filter(contentartifact__pk__in=content_artifacts)
        repo_versions = RepositoryVersion.objects.with_content(content).select_related("repository")

        self.stdout.write(
            _("Found {} on-demand content units with forbidden checksums.").format(content.count())
        )
        if content.count():
            self.stdout.write(
                _("There is approx {:.2f}Mb of content to be downloaded.").format(
                    ras_size / (1024**2)
                )
            )

        if repo_versions.exists():
            self.stdout.write(_("\nAffected repository versions with remote content:"))
            self._print_out_repository_version_hrefs(repo_versions)

    def _show_immediate_content(self, forbidden_checksums):
        allowed_checksums = set(
            constants.ALL_KNOWN_CONTENT_CHECKSUMS.symmetric_difference(forbidden_checksums)
        )
        query_forbidden = Q()
        query_required = Q()
        for checksum in forbidden_checksums:
            query_forbidden |= Q(**{f"{checksum}__isnull": False})

        for allowed_checksum in allowed_checksums:
            query_required |= Q(**{f"{allowed_checksum}__isnull": True})

        artifacts = Artifact.objects.filter(query_forbidden | query_required)
        content_artifacts = ContentArtifact.objects.filter(artifact__in=artifacts)
        content = Content.objects.filter(contentartifact__pk__in=content_artifacts)
        repo_versions = RepositoryVersion.objects.with_content(content).select_related("repository")

        self.stdout.write(
            _("Found {} downloaded content units with forbidden or missing checksums.").format(
                content.count()
            )
        )
        if content.count() > 0:
            self.stdout.write(
                _("There is approx. {:.2f}Mb content data to be re-hashed.").format(
                    artifacts.aggregate(Sum("size"))["size__sum"] / (1024**2)
                )
            )

        if repo_versions.exists():
            self.stdout.write(_("\nAffected repository versions with present content:"))
            self._print_out_repository_version_hrefs(repo_versions)

    def _download_artifact(self, artifact, checksum, file_path):
        restored = False
        for ca in artifact.content_memberships.all():
            if not restored:
                for ra in ca.remoteartifact_set.all():
                    remote = ra.remote.cast()
                    if remote.policy == "immediate":
                        self.stdout.write(_("Restoring missing file {}").format(file_path))
                        downloader = remote.get_downloader(ra)
                        dl_result = downloader.fetch()
                        # FIXME in case url is not available anymore this will break
                        if dl_result.artifact_attributes["sha256"] == artifact.sha256:
                            with open(dl_result.path, "rb") as src:
                                filename = artifact.file.name
                                artifact.file.save(filename, src, save=False)
                            setattr(artifact, checksum, dl_result.artifact_attributes[checksum])
                            restored = True
                            break
            else:
                break
        return restored

    def _report(self, allowed_checksums):
        if allowed_checksums:
            allowed_checksums = allowed_checksums.split(",")
            if "sha256" not in allowed_checksums:
                raise CommandError(_("Checksums must contain sha256"))
        else:
            allowed_checksums = settings.ALLOWED_CONTENT_CHECKSUMS

        forbidden_checksums = set(constants.ALL_KNOWN_CONTENT_CHECKSUMS).difference(
            allowed_checksums
        )

        self.stderr.write(
            _(
                "Warning: the handle-artifact-checksums report is in "
                "tech preview and may change in the future."
            )
        )
        self._show_on_demand_content(forbidden_checksums)
        self._show_immediate_content(forbidden_checksums)

    def handle(self, *args, **options):

        if options["report"]:
            self._report(options["checksums"])
            return

        if options["checksums"] and not options["report"]:
            self.stdout.write(_("Checksums cannot be supplied without --report argument"))
            exit(1)

        log.setLevel(logging.ERROR)
        hrefs = set()
        for checksum in settings.ALLOWED_CONTENT_CHECKSUMS:
            params = {f"{checksum}__isnull": True}
            artifacts_qs = Artifact.objects.filter(**params)
            artifacts = []
            for a in artifacts_qs.iterator():
                hasher = pulp_hashlib.new(checksum)
                try:
                    with a.file as fp:
                        for chunk in fp.chunks(CHUNK_SIZE):
                            hasher.update(chunk)
                    setattr(a, checksum, hasher.hexdigest())
                except FileNotFoundError:
                    file_path = os.path.join(settings.MEDIA_ROOT, a.file.name)
                    restored = self._download_artifact(a, checksum, file_path)
                    if not restored:
                        hrefs.add(file_path)
                artifacts.append(a)

                if len(artifacts) >= 1000:
                    Artifact.objects.bulk_update(objs=artifacts, fields=[checksum], batch_size=1000)
                    artifacts.clear()

            if artifacts:
                Artifact.objects.bulk_update(objs=artifacts, fields=[checksum])

        if hrefs:
            raise CommandError(
                _("Some files that were missing could not be restored: {}").format(hrefs)
            )

        forbidden_checksums = set(constants.ALL_KNOWN_CONTENT_CHECKSUMS).difference(
            settings.ALLOWED_CONTENT_CHECKSUMS
        )
        for checksum in forbidden_checksums:
            search_params = {f"{checksum}__isnull": False}
            update_params = {f"{checksum}": None}
            artifacts_qs = Artifact.objects.filter(**search_params)
            if artifacts_qs.exists():
                self.stdout.write(
                    _("Removing forbidden checksum {} from database").format(checksum)
                )
                artifacts_qs.update(**update_params)

        self.stdout.write(_("Finished aligning checksums with settings.ALLOWED_CONTENT_CHECKSUMS"))



from itertools import compress
from uuid import uuid4

from django.test import TestCase
from pulpcore.plugin.models import Content, Label, Repository, RepositoryVersion


class RepositoryVersionTestCase(TestCase):
    def setUp(self):
        self.repository = Repository.objects.create()
        self.repository.CONTENT_TYPES = [Content]
        self.repository.save()

        contents = []
        for _ in range(0, 5):
            contents.append(Content(pulp_type="core.content"))

        Content.objects.bulk_create(contents)
        self.pks = [c.pk for c in contents]

    def test_add_and_remove_content(self):
        contents = Content.objects.filter(pk__in=self.pks[:4])
        with self.repository.new_version() as version1:
            version1.add_content(contents)  # v1 == four content units 0-3

        to_remove = contents[0:2]
        with self.repository.new_version() as version2:
            version2.remove_content(to_remove)  # v2 == removed 2 first contents

        to_add = Content.objects.filter(pk=contents[0].pk)
        with self.repository.new_version() as version3:
            version3.add_content(to_add)  # v3 == added first content

        version0 = RepositoryVersion.objects.filter(number=0, repository=self.repository).first()

        self.assertEqual(version0.added().count(), 0)
        self.assertEqual(version1.added().count(), 4)
        self.assertEqual(version2.added().count(), 0)
        self.assertEqual(version3.added().count(), 1)

        self.assertEqual(version0.removed().count(), 0)
        self.assertEqual(version1.removed().count(), 0)
        self.assertEqual(version2.removed().count(), 2)
        self.assertEqual(version3.removed().count(), 0)

        self.assertEqual(version3.added(version0).count(), 3)
        self.assertEqual(version3.removed(version0).count(), 0)

        added_pks_0 = version3.added(version0).values_list("pk", flat=True)
        removed_pks_0 = version3.removed(version0).values_list("pk", flat=True)

        self.assertCountEqual(added_pks_0, compress(self.pks, [1, 0, 1, 1]), added_pks_0)
        self.assertCountEqual(removed_pks_0, compress(self.pks, [0, 0, 0, 0]), removed_pks_0)

        added_pks_1 = version3.added(version1).values_list("pk", flat=True)
        removed_pks_1 = version3.removed(version1).values_list("pk", flat=True)

        self.assertCountEqual(added_pks_1, compress(self.pks, [0, 0, 0, 0]), added_pks_1)
        self.assertCountEqual(removed_pks_1, compress(self.pks, [0, 1, 0, 0]), removed_pks_1)

        added_pks_2 = version3.added(version2).values_list("pk", flat=True)
        removed_pks_2 = version3.removed(version2).values_list("pk", flat=True)

        self.assertCountEqual(added_pks_2, compress(self.pks, [1, 0, 0, 0]), added_pks_2)
        self.assertCountEqual(removed_pks_2, compress(self.pks, [0, 0, 0, 0]), removed_pks_2)

    def content_qs(self, pks):
        return Content.objects.filter(pk__in=pks)

    def verify_content_sets(self, version, content, added, removed):
        """
        Verify the content, added, and removed sets for a repository version.

        Args:
            version (pulpcore.app.models.RepositoryVersion): the version instance to verify
            content (list): "presence list" for content with respect to `self.pks`.
                For example, [1, 0, 1] means that content with self.pks[0], and
                self.pks[2] must be present, all other content must not be present.
            added (list): "presence list" for added content
            remove (list): "presence list" for removed content

        """
        content_pks = version.content.values_list("pk", flat=True)
        added_pks = version.added().values_list("pk", flat=True)
        removed_pks = version.removed().values_list("pk", flat=True)

        # There must never be content shown as added & removed
        self.assertSetEqual(set(added_pks).intersection(removed_pks), set())

        self.assertCountEqual(content_pks, compress(self.pks, content), content_pks)
        self.assertCountEqual(added_pks, compress(self.pks, added), added_pks)
        self.assertCountEqual(removed_pks, compress(self.pks, removed), removed_pks)

    def test_add_remove(self):
        """Verify that adding and then removing content units is handled properly."""
        latest_version = self.repository.latest_version()

        with self.repository.new_version() as version1:
            version1.add_content(self.content_qs(self.pks[:5]))
            self.verify_content_sets(version1, content=[1] * 5, added=[1] * 5, removed=[])

            version1.remove_content(self.content_qs(self.pks[:5]))
            self.verify_content_sets(version1, content=[], added=[], removed=[])

        self.assertEqual(
            self.repository.latest_version(), latest_version, msg="Empty version1 must not exist."
        )

    def test_remove_add(self):
        """Verify that removing and then adding content units is handled properly."""
        with self.repository.new_version() as version1:
            version1.add_content(self.content_qs(self.pks[:5]))  # v1 == content ids 0-4

        with self.repository.new_version() as version2:
            version2.remove_content(self.content_qs(self.pks[:5]))
            self.verify_content_sets(version2, content=[], added=[], removed=[1] * 5)

            version2.add_content(self.content_qs(self.pks[:5]))
            self.verify_content_sets(version2, content=[1] * 5, added=[], removed=[])

        self.assertEqual(
            self.repository.latest_version(), version1, msg="Empty version2 must not exist."
        )

    def test_multiple_adds_and_removes(self):
        """Verify that adding/removing content multiple times is handled properly.

        Additionally, verify that other content (untouched, simple add, simple
        remove) is not influenced and behaves as expected.
        """
        # v1 == content id 0, 2, and 4
        with self.repository.new_version() as version1:
            version1.add_content(self.content_qs([self.pks[0], self.pks[2], self.pks[4]]))

        # v2 version is created in multiple steps:
        #
        # |       | content id                                 |
        # |       | 0      | 1      | 2      | 3      | 4      |
        # | step1 |        |        | remove |        | remove |
        # | step2 |        | add    |        | add    | add    |
        # | step3 |        |        |        | remove | remove |
        # | step4 |        |        |        | add    |        |
        #
        # Expected outcome after step 4:
        # content: 0, 1, 3
        # added: 1, 3
        # removed: 2, 4
        with self.repository.new_version() as version2:
            # Content must be that of version1:
            self.verify_content_sets(version2, content=[1, 0, 1, 0, 1], added=[], removed=[])

            # step 1
            version2.remove_content(self.content_qs([self.pks[2], self.pks[4]]))
            self.verify_content_sets(
                version2, content=[1, 0, 0, 0, 0], added=[], removed=[0, 0, 1, 0, 1]
            )

            # step 2
            version2.add_content(self.content_qs([self.pks[1], self.pks[3], self.pks[4]]))

            self.verify_content_sets(
                version2, content=[1, 1, 0, 1, 1], added=[0, 1, 0, 1, 0], removed=[0, 0, 1, 0, 0]
            )

            # step 3
            version2.remove_content(self.content_qs([self.pks[3], self.pks[4]]))

            self.verify_content_sets(
                version2, content=[1, 1, 0, 0, 0], added=[0, 1, 0, 0, 0], removed=[0, 0, 1, 0, 1]
            )

            # step 4
            version2.add_content(self.content_qs([self.pks[3]]))

            self.verify_content_sets(
                version2, content=[1, 1, 0, 1, 0], added=[0, 1, 0, 1, 0], removed=[0, 0, 1, 0, 1]
            )

        # Verify content sets after saving
        self.verify_content_sets(
            version1, content=[1, 0, 1, 0, 1, 0], added=[1, 0, 1, 0, 1, 0], removed=[]
        )
        self.verify_content_sets(
            version2, content=[1, 1, 0, 1, 0], added=[0, 1, 0, 1, 0], removed=[0, 0, 1, 0, 1]
        )

    @staticmethod
    def pks_of_next_qs(qs_generator):
        """Iterate qs_generator one step and return the list of pks in the qs."""
        return list(next(qs_generator).values_list("pk", flat=True))

    def test_content_batch_qs(self):
        """Verify content iteration using content_batch_qs()."""
        sorted_pks = sorted(self.pks[:4])
        contents = Content.objects.filter(pk__in=self.pks[:4])
        with self.repository.new_version() as version1:
            version1.add_content(contents)  # v1 == four content units 0-3

            # Verify content_batch_qs on incomplete version
            qs_generator = version1.content_batch_qs(batch_size=2)
            self.assertListEqual(self.pks_of_next_qs(qs_generator), sorted_pks[:2])
            self.assertListEqual(self.pks_of_next_qs(qs_generator), sorted_pks[2:])
            with self.assertRaises(StopIteration):
                self.pks_of_next_qs(qs_generator)

        # Verify on complete version
        # The last batch has only a single element (Depending on how a qs is
        # written, Django ORM sometimes returns the actual object instead of a
        # qs. This must not happen for the generator).
        reverse_pks = sorted(self.pks[:4], reverse=True)
        qs_generator = version1.content_batch_qs(order_by_params=("-pk",), batch_size=3)
        self.assertListEqual(self.pks_of_next_qs(qs_generator), reverse_pks[:3])
        self.assertListEqual(self.pks_of_next_qs(qs_generator), reverse_pks[3:])
        with self.assertRaises(StopIteration):
            self.pks_of_next_qs(qs_generator)

    def test_content_batch_qs_using_filter(self):
        """Verify that a plugin can define a filtering query set for content_batch_qs()."""
        contents = Content.objects.filter(pk__in=self.pks[:4])
        with self.repository.new_version() as version1:
            version1.add_content(contents)  # v1 == four content units 0-3

        # Filter for pks 0-2 and 4 (4 is not part of the repo version, i.e.
        # must not be part of the content iterated over).
        pre_filter_qs = Content.objects.filter(pk__in=self.pks[:3] + [self.pks[4]])

        sorted_pks = sorted(self.pks[:3])
        qs_generator = version1.content_batch_qs(content_qs=pre_filter_qs, batch_size=2)
        self.assertListEqual(self.pks_of_next_qs(qs_generator), sorted_pks[:2])
        self.assertListEqual(self.pks_of_next_qs(qs_generator), sorted_pks[2:])
        with self.assertRaises(StopIteration):
            self.pks_of_next_qs(qs_generator)


class RepositoryTestCase(TestCase):
    def setUp(self):
        self.repository = Repository.objects.create()
        self.repository.CONTENT_TYPES = [Content]
        self.repository.save()

    def test_next_version_with_one_version(self):
        self.assertEqual(self.repository.next_version, 1, self.repository.next_version)
        self.assertEqual(
            self.repository.latest_version().number, 0, self.repository.latest_version().number
        )
        content = Content(pulp_type="core.content")
        content.save()

        with self.repository.new_version() as version:
            version.add_content(Content.objects.filter(pk=content.pk))

        self.assertEqual(self.repository.next_version, 2, self.repository.next_version)
        self.assertEqual(
            self.repository.latest_version().number, 1, self.repository.latest_version().number
        )

        version.delete()

        self.assertEqual(self.repository.next_version, 2, self.repository.next_version)
        self.assertEqual(
            self.repository.latest_version().number, 0, self.repository.latest_version().number
        )

    def test_next_version_with_multiple_versions(self):
        self.assertEqual(self.repository.next_version, 1, self.repository.next_version)
        self.assertEqual(
            self.repository.latest_version().number, 0, self.repository.latest_version().number
        )
        contents = []
        for _ in range(0, 3):
            contents.append(Content(pulp_type="core.content"))

        Content.objects.bulk_create(contents)

        versions = [self.repository.latest_version()]
        for content in contents:
            with self.repository.new_version() as version:
                version.add_content(Content.objects.filter(pk=content.pk))
                versions.append(version)

        self.assertEqual(self.repository.next_version, 4, self.repository.next_version)
        self.assertEqual(
            self.repository.latest_version().number, 3, self.repository.latest_version().number
        )

        versions[2].delete()
        versions[3].delete()

        self.assertEqual(self.repository.next_version, 4, self.repository.next_version)
        self.assertEqual(
            self.repository.latest_version().number, 1, self.repository.latest_version().number
        )

    def test_label_cascade(self):
        """Check that a repo's labels are deleted when it's deleted."""
        repo = Repository.objects.create(name=uuid4())
        repo.CONTENT_TYPES = [Content]
        repo.save()

        label = repo.pulp_labels.create(key="brown", value="radagast")
        self.assertTrue(Label.objects.filter(pk=label.pk).exists())
        repo.delete()
        self.assertFalse(Label.objects.filter(pk=label.pk).exists())



import logging
from gettext import gettext as _

from django.db import transaction
from django.db import connection
from django.urls import reverse

from pulpcore.app.models import Task
from pulpcore.app.util import get_view_name_for_model
from pulpcore.constants import TASK_FINAL_STATES, TASK_INCOMPLETE_STATES, TASK_STATES
from pulpcore.exceptions import MissingResource

_logger = logging.getLogger(__name__)


def cancel(task_id):
    """
    Cancel the task that is represented by the given task_id.

    This method cancels only the task with given task_id, not the spawned tasks. This also updates
    task's state to either 'canceled' or 'canceling'.

    Args:
        task_id (str): The ID of the task you wish to cancel

    Raises:
        MissingResource: if a task with given task_id does not exist
    """
    try:
        task_status = Task.objects.get(pk=task_id)
    except Task.DoesNotExist:
        raise MissingResource(task=task_id)

    if task_status.state in TASK_FINAL_STATES:
        # If the task is already done, just stop
        msg = _("Task [{task_id}] already in a final state: {state}")
        _logger.debug(msg.format(task_id=task_id, state=task_status.state))
        return task_status

    _logger.info(_("Canceling task: {id}").format(id=task_id))

    task = task_status
    # This is the only valid transition without holding the task lock
    rows = Task.objects.filter(pk=task.pk, state__in=TASK_INCOMPLETE_STATES).update(
        state=TASK_STATES.CANCELING
    )
    # Notify the worker that might be running that task and other workers to clean up
    with connection.cursor() as cursor:
        cursor.execute("SELECT pg_notify('pulp_worker_cancel', %s)", (str(task.pk),))
        cursor.execute("NOTIFY pulp_worker_wakeup")
    if rows == 1:
        task.refresh_from_db()
    return task


def _delete_incomplete_resources(task):
    """
    Delete all incomplete created-resources on a canceled task.

    Args:
        task (Task): A task.
    """
    if task.state not in [TASK_STATES.CANCELED, TASK_STATES.CANCELING]:
        raise RuntimeError(_("Task must be canceled."))
    for model in (r.content_object for r in task.created_resources.all()):
        try:
            if model.complete:
                continue
        except AttributeError:
            continue
        try:
            with transaction.atomic():
                model.delete()
        except Exception as error:
            _logger.error(_("Delete created resource, failed: {}").format(str(error)))


def get_url(model):
    """
    Get a resource url for the specified model object. This returns the path component of the
    resource URI.  This is used in our resource locking/reservation code to identify resources.

    Args:
        model (django.models.Model): A model object.

    Returns:
        str: The path component of the resource url
    """
    return reverse(get_view_name_for_model(model, "detail"), args=[model.pk])



# Generated by Django 2.2.19 on 2021-03-05 18:35
import logging
import os
from gettext import gettext as _

from django.conf import settings
from django.db import migrations

from pulpcore import constants
from pulpcore.app import pulp_hashlib

CHUNK_SIZE = 1024 * 1024  # 1 Mb

_logger = logging.getLogger(__name__)


def handle_artifact_checksums(apps):
    Artifact = apps.get_model("core", "Artifact")
    paths = []

    # set allowed checksums
    for checksum in settings.ALLOWED_CONTENT_CHECKSUMS:
        params = {f"{checksum}__isnull": True}
        artifacts_qs = Artifact.objects.filter(**params)
        artifacts = []
        for a in artifacts_qs.iterator():
            hasher = pulp_hashlib.new(checksum)
            try:
                for chunk in a.file.chunks(CHUNK_SIZE):
                    hasher.update(chunk)
                a.file.close()
                setattr(a, checksum, hasher.hexdigest())
                artifacts.append(a)
            except FileNotFoundError:
                file_path = os.path.join(settings.MEDIA_ROOT, a.file.name)
                paths.append(file_path)

            if len(artifacts) >= 1000:
                Artifact.objects.bulk_update(objs=artifacts, fields=[checksum], batch_size=1000)
                artifacts.clear()

        if artifacts:
            Artifact.objects.bulk_update(objs=artifacts, fields=[checksum], batch_size=1000)

        if paths:
            _logger.warn(
                _(
                    "Missing files needed to update artifact {checksum} checksum: {paths}. "
                    "Please run 'pulpcore-manager handle-artifact-checksums'."
                ).format(checksum=checksum, paths=paths)
            )

        # unset any forbidden checksums
        forbidden_checksums = set(constants.ALL_KNOWN_CONTENT_CHECKSUMS).difference(
            settings.ALLOWED_CONTENT_CHECKSUMS
        )
        for checksum in forbidden_checksums:
            search_params = {f"{checksum}__isnull": False}
            update_params = {f"{checksum}": None}
            artifacts_qs = Artifact.objects.filter(**search_params)
            if artifacts_qs.exists():
                artifacts_qs.update(**update_params)


def run_handle_artifact_checksums(apps, schema_editor):
    try:
        handle_artifact_checksums(apps)
    except Exception as exc:
        _logger.warn(
            _(
                "Failed to update checksums for artifacts: {}. "
                "Please run 'pulpcore-manager handle-artifact-checksums'."
            ).format(exc)
        )


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0060_data_migration_proxy_creds"),
    ]

    operations = [
        migrations.RunPython(
            run_handle_artifact_checksums, reverse_code=run_handle_artifact_checksums
        ),
    ]



from gettext import gettext as _

import asyncio
from collections import namedtuple
import logging
import os
import tempfile

from pulpcore.app import pulp_hashlib
from pulpcore.app.models import Artifact
from pulpcore.exceptions import (
    DigestValidationError,
    SizeValidationError,
    TimeoutException,
    UnsupportedDigestValidationError,
)


log = logging.getLogger(__name__)


DownloadResult = namedtuple("DownloadResult", ["url", "artifact_attributes", "path", "headers"])
"""
Args:
    url (str): The url corresponding with the download.
    path (str): The absolute path to the saved file
    artifact_attributes (dict): Contains keys corresponding with
        :class:`~pulpcore.plugin.models.Artifact` fields. This includes the computed digest values
        along with size information.
    headers (aiohttp.multidict.MultiDict): HTTP response headers. The keys are header names. The
        values are header content. None when not using the HttpDownloader or sublclass.
"""


class BaseDownloader:
    """
    The base class of all downloaders, providing digest calculation, validation, and file handling.

    This is an abstract class and is meant to be subclassed. Subclasses are required to implement
    the :meth:`~pulpcore.plugin.download.BaseDownloader.run` method and do two things:

        1. Pass all downloaded data to
           :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data` and schedule it.

        2. Schedule :meth:`~pulpcore.plugin.download.BaseDownloader.finalize` after all data has
           been delivered to :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.

    Passing all downloaded data the into
    :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data` allows the file digests to
    be computed while data is written to disk. The digests computed are required if the download is
    to be saved as an :class:`~pulpcore.plugin.models.Artifact` which avoids having to re-read the
    data later.

    The :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data` method by default
    writes to a random file in the current working directory or you can pass in your own file
    object. See the ``custom_file_object`` keyword argument for more details. Allowing the download
    instantiator to define the file to receive data allows the streamer to receive the data instead
    of having it written to disk.

    The call to :meth:`~pulpcore.plugin.download.BaseDownloader.finalize` ensures that all
    data written to the file-like object is quiesced to disk before the file-like object has
    `close()` called on it.

    Attributes:
        url (str): The url to download.
        expected_digests (dict): Keyed on the algorithm name provided by hashlib and stores the
            value of the expected digest. e.g. {'md5': '912ec803b2ce49e4a541068d495ab570'}
        expected_size (int): The number of bytes the download is expected to have.
        path (str): The full path to the file containing the downloaded data if no
            ``custom_file_object`` option was specified, otherwise None.
    """

    def __init__(
        self,
        url,
        custom_file_object=None,
        expected_digests=None,
        expected_size=None,
        semaphore=None,
        *args,
        **kwargs,
    ):
        """
        Create a BaseDownloader object. This is expected to be called by all subclasses.

        Args:
            url (str): The url to download.
            custom_file_object (file object): An open, writable file object that downloaded data
                can be written to by
                :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.
            expected_digests (dict): Keyed on the algorithm name provided by hashlib and stores the
                value of the expected digest. e.g. {'md5': '912ec803b2ce49e4a541068d495ab570'}
            expected_size (int): The number of bytes the download is expected to have.
            semaphore (asyncio.Semaphore): A semaphore the downloader must acquire before running.
                Useful for limiting the number of outstanding downloaders in various ways.
        """
        self.url = url
        self._writer = custom_file_object
        self.path = None
        self.expected_digests = expected_digests
        self.expected_size = expected_size
        if semaphore:
            self.semaphore = semaphore
        else:
            self.semaphore = asyncio.Semaphore()  # This will always be acquired
        self._digests = {}
        self._size = 0
        if self.expected_digests:
            if not set(self.expected_digests).intersection(set(Artifact.DIGEST_FIELDS)):
                raise UnsupportedDigestValidationError(
                    _(
                        "Content at the url {} does not contain at least one trusted hasher which"
                        " is specified in 'ALLOWED_CONTENT_CHECKSUMS' setting."
                    ).format(self.url)
                )

    def _ensure_writer_has_open_file(self):
        """
        Create a temporary file on demand.

        Create a temporary file when it's actually used,
        allowing plugin writers to instantiate many downloaders in memory.
        """
        if not self._writer:
            self._writer = tempfile.NamedTemporaryFile(dir=os.getcwd(), delete=False)
            self.path = self._writer.name
            self._digests = {n: pulp_hashlib.new(n) for n in Artifact.DIGEST_FIELDS}
            self._size = 0

    async def handle_data(self, data):
        """
        A coroutine that writes data to the file object and compute its digests.

        All subclassed downloaders are expected to pass all data downloaded to this method. Similar
        to the hashlib docstring, repeated calls are equivalent to a single call with
        the concatenation of all the arguments: m.handle_data(a); m.handle_data(b) is equivalent to
        m.handle_data(a+b).

        Args:
            data (bytes): The data to be handled by the downloader.
        """
        self._ensure_writer_has_open_file()
        self._writer.write(data)
        self._record_size_and_digests_for_data(data)

    async def finalize(self):
        """
        A coroutine to flush downloaded data, close the file writer, and validate the data.

        All subclasses are required to call this method after all data has been passed to
        :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.

        Raises:
            :class:`~pulpcore.exceptions.DigestValidationError`: When any of the ``expected_digest``
                values don't match the digest of the data passed to
                :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.
            :class:`~pulpcore.exceptions.SizeValidationError`: When the ``expected_size`` value
                doesn't match the size of the data passed to
                :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.
        """
        self._ensure_writer_has_open_file()
        self._writer.flush()
        os.fsync(self._writer.fileno())
        self._writer.close()
        self._writer = None
        self.validate_digests()
        self.validate_size()

    def fetch(self):
        """
        Run the download synchronously and return the `DownloadResult`.

        Returns:
            :class:`~pulpcore.plugin.download.DownloadResult`

        Raises:
            Exception: Any fatal exception emitted during downloading
        """
        done, _ = asyncio.get_event_loop().run_until_complete(asyncio.wait([self.run()]))
        return done.pop().result()

    def _record_size_and_digests_for_data(self, data):
        """
        Record the size and digest for an available chunk of data.

        Args:
            data (bytes): The data to have its size and digest values recorded.
        """
        for algorithm in self._digests.values():
            algorithm.update(data)
        self._size += len(data)

    @property
    def artifact_attributes(self):
        """
        A property that returns a dictionary with size and digest information. The keys of this
        dictionary correspond with :class:`~pulpcore.plugin.models.Artifact` fields.
        """
        attributes = {"size": self._size}
        for algorithm in self._digests:
            attributes[algorithm] = self._digests[algorithm].hexdigest()
        return attributes

    def validate_digests(self):
        """
        Validate all digests validate if ``expected_digests`` is set

        Raises:
            :class:`~pulpcore.exceptions.DigestValidationError`: When any of the ``expected_digest``
                values don't match the digest of the data passed to
                :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.
        """
        if self.expected_digests:
            for algorithm, expected_digest in self.expected_digests.items():
                if expected_digest != self._digests[algorithm].hexdigest():
                    raise DigestValidationError(self.url)

    def validate_size(self):
        """
        Validate the size if ``expected_size`` is set

        Raises:
            :class:`~pulpcore.exceptions.SizeValidationError`: When the ``expected_size`` value
                doesn't match the size of the data passed to
                :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.
        """
        if self.expected_size:
            if self._size != self.expected_size:
                raise SizeValidationError(self.url)

    async def run(self, extra_data=None):
        """
        Run the downloader with concurrency restriction.

        This method acquires `self.semaphore` before calling the actual download implementation
        contained in `_run()`. This ensures that the semaphore stays acquired even as the `backoff`
        decorator on `_run()`, handles backoff-and-retry logic.

        Args:
            extra_data (dict): Extra data passed to the downloader.

        Returns:
            :class:`~pulpcore.plugin.download.DownloadResult` from `_run()`.

        """
        async with self.semaphore:
            try:
                return await self._run(extra_data=extra_data)
            except asyncio.TimeoutError:
                raise TimeoutException(self.url)

    async def _run(self, extra_data=None):
        """
        Run the downloader.

        This is a coroutine that asyncio can schedule to complete downloading. Subclasses are
        required to implement this method and do two things:

        1. Pass all downloaded data to
           :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.

        2. Call :meth:`~pulpcore.plugin.download.BaseDownloader.finalize` after all data has
           been delivered to :meth:`~pulpcore.plugin.download.BaseDownloader.handle_data`.

        It is also expected that the subclass implementation return a
        :class:`~pulpcore.plugin.download.DownloadResult` object. The
        ``artifact_attributes`` value of the
        :class:`~pulpcore.plugin.download.DownloadResult` is usually set to the
        :attr:`~pulpcore.plugin.download.BaseDownloader.artifact_attributes` property value.

        This method is called from :meth:`~pulpcore.plugin.download.BaseDownloader.run` which
        handles concurrency restriction. Thus, by the time this method is called, the download can
        occur without violating the concurrency restriction.

        Args:
            extra_data (dict): Extra data passed to the downloader.

        Returns:
            :class:`~pulpcore.plugin.download.DownloadResult`

        Raises:
            Validation errors could be emitted when subclassed implementations call
            :meth:`~pulpcore.plugin.download.BaseDownloader.finalize`.
        """
        raise NotImplementedError("Subclasses must define a _run() method that returns a coroutine")



# Generated by Django 2.2.5 on 2019-09-09 20:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_add_duplicated_reserved_resources'),
    ]

    operations = [
        migrations.AddField(
            model_name='progressreport',
            name='code',
            field=models.CharField(default="no_code", max_length=36),
            preserve_default=False,
        ),
    ]



"""Tests that perform actions over reclaim disk space."""
from pulp_smash import config
from pulp_smash.pulp3.bindings import monitor_task, PulpTestCase
from pulp_smash.pulp3.utils import (
    gen_repo,
    get_content,
    gen_distribution,
    download_content_unit,
)
from pulpcore.client.pulpcore import (
    ArtifactsApi,
    OrphansCleanupApi,
    RepositoriesReclaimSpaceApi,
)
from pulpcore.client.pulp_file import (
    FileFilePublication,
    PublicationsFileApi,
    RepositoriesFileApi,
    RepositorySyncURL,
    RemotesFileApi,
    DistributionsFileApi,
)
from pulpcore.tests.functional.api.using_plugin.constants import FILE_CONTENT_NAME
from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
    gen_file_remote,
)
from pulpcore.tests.functional.utils import core_client


class ReclaimSpaceTestCase(PulpTestCase):
    """
    Test whether repository content can be reclaimed.
    Subsequently, confirm that artifact is correctly re-downloaded in sync
    task or when streamed to the client (this is true only for synced content, not uploaded.)
    """

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = gen_file_client()
        cls.orphans_api = OrphansCleanupApi(core_client)
        cls.reclaim_api = RepositoriesReclaimSpaceApi(core_client)
        cls.artifacts_api = ArtifactsApi(core_client)
        cls.publication_api = PublicationsFileApi(cls.client)
        cls.distributions_api = DistributionsFileApi(cls.client)
        cls.repo_api = RepositoriesFileApi(cls.client)
        cls.remote_api = RemotesFileApi(cls.client)

        orphans_response = cls.orphans_api.cleanup({"orphan_protection_time": 0})
        monitor_task(orphans_response.task)

    def tearDown(self):
        """Clean created resources."""
        # Runs any delete tasks and waits for them to complete
        self.doCleanups()
        orphans_response = self.orphans_api.cleanup({"orphan_protection_time": 0})
        monitor_task(orphans_response.task)

    def test_reclaim_immediate_content(self):
        """
        Test whether immediate repository content can be reclaimed
        and then re-populated back after sync.
        """
        repo = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, repo.pulp_href)

        remote = self.remote_api.create(gen_file_remote())
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        # sync the repository with immediate policy
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        # reclaim disk space
        reclaim_response = self.reclaim_api.reclaim({"repo_hrefs": [repo.pulp_href]})
        monitor_task(reclaim_response.task)

        # assert no artifacts left
        artifacts = self.artifacts_api.list().count
        self.assertEqual(artifacts, 0)

        # sync repo again
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        # assert re-sync populated missing artifacts
        artifacts = self.artifacts_api.list().count
        self.assertGreater(artifacts, 0)
        self.addCleanup(self.orphans_api.cleanup, {"orphan_protection_time": 0})

    def test_reclaim_on_demand_content(self):
        """
        Test whether on_demand repository content can be reclaimed
        and then re-populated back after client request.
        """
        repo = self.repo_api.create(gen_repo())
        self.addCleanup(self.repo_api.delete, repo.pulp_href)

        # sync the repository with on_demand policy
        body = gen_file_remote(**{"policy": "on_demand"})
        remote = self.remote_api.create(body)
        self.addCleanup(self.remote_api.delete, remote.pulp_href)

        # sync repo
        repository_sync_data = RepositorySyncURL(remote=remote.pulp_href)
        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = self.repo_api.read(repo.pulp_href)

        # Publication
        publication_data = FileFilePublication(repository=repo.pulp_href)
        publication_response = self.publication_api.create(publication_data)
        task_response = monitor_task(publication_response.task)
        publication = self.publication_api.read(task_response.created_resources[0])
        self.addCleanup(self.publication_api.delete, publication.pulp_href)

        # Distribution
        body = gen_distribution()
        body["publication"] = publication.pulp_href
        distribution_response = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_response.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])
        self.addCleanup(self.distributions_api.delete, distribution.pulp_href)

        artifacts_before_download = self.artifacts_api.list().count
        content = get_content(repo.to_dict())[FILE_CONTENT_NAME][0]
        download_content_unit(self.cfg, distribution.to_dict(), content["relative_path"])

        artifacts = self.artifacts_api.list().count
        self.assertGreater(artifacts, artifacts_before_download)

        # reclaim disk space
        reclaim_response = self.reclaim_api.reclaim({"repo_hrefs": [repo.pulp_href]})
        monitor_task(reclaim_response.task)

        artifacts_after_reclaim = self.artifacts_api.list().count
        content = get_content(repo.to_dict())[FILE_CONTENT_NAME]
        download_content_unit(self.cfg, distribution.to_dict(), content[0]["relative_path"])

        artifacts = self.artifacts_api.list().count
        self.assertGreater(artifacts, artifacts_after_reclaim)



from rest_framework.response import Response
from rest_framework.reverse import reverse


class OperationPostponedResponse(Response):
    """
    An HTTP response class for returning 202 and a spawned task.

    This response object should be used by views that dispatch asynchronous tasks. The most common
    use case is for sync and publish operations. When JSON is requested, the response will look
    like the following::

        {
            "task": "/pulp/api/v3/tasks/735633bc-eb41-4737-b436-c7c6914f34b1/"
        }
    """

    def __init__(self, task, request):
        """
        Args:
            task (pulpcore.plugin.models.Task): A :class:`~pulpcore.plugin.models.Task` object
                used to generate the response.
            request (rest_framework.request.Request): Request used to generate the pulp_href urls
        """
        resp = {"task": reverse("tasks-detail", args=[task.pk], request=None)}
        super().__init__(data=resp, status=202)


class TaskGroupOperationResponse(Response):
    """
    An HTTP response class for returning 202 and a task group.

    This response object should be used by views that create a task group and dispatch one or more
    tasks for that group. When JSON is requested, the response will look like the following::

        {
            "task_group": "/pulp/api/v3/task-groups/735633bc-eb41-4737-b436-c7c6914f34b1/"
        }
    """

    def __init__(self, task_group, request):
        """
        Args:
            task_group (pulpcore.plugin.models.TaskGroup): A
                :class:`~pulpcore.plugin.models.TaskGroup` object used to generate the response.
            request (rest_framework.request.Request): Request used to generate the pulp_href urls
        """
        resp = {"task_group": reverse("task-groups-detail", args=[task_group.pk], request=None)}
        super().__init__(data=resp, status=202)



from uuid import uuid4
from unittest.mock import patch, mock_open

from django.db import connection
from django.test import TestCase

from pulpcore.app.models import Remote
from pulpcore.app.models.fields import EncryptedTextField


class RemoteTestCase(TestCase):
    def setUp(self):
        self.remote = None

    def tearDown(self):
        if self.remote:
            self.remote.delete()

    @patch(
        "pulpcore.app.models.fields.open",
        new_callable=mock_open,
        read_data=b"hPCIFQV/upbvPRsEpgS7W32XdFA2EQgXnMtyNAekebQ=",
    )
    def test_encrypted_proxy_password(self, mock_file):
        self.remote = Remote(name=uuid4(), proxy_password="test")
        self.remote.save()
        self.assertEqual(Remote.objects.get(pk=self.remote.pk).proxy_password, "test")

        # check the database that proxy_password is encrypted
        with connection.cursor() as cursor:
            cursor.execute(
                "SELECT proxy_password FROM core_remote " f"WHERE pulp_id = '{self.remote.pulp_id}'"
            )
            db_proxy_password = cursor.fetchone()[0]
            proxy_password = EncryptedTextField().from_db_value(db_proxy_password, None, connection)
            self.assertNotEqual(db_proxy_password, "test")
            self.assertEqual(proxy_password, "test")
            self.assertEqual(mock_file.call_count, 2)



# Generated by Django 2.2.11 on 2020-03-31 17:49

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0025_task_parent_task'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaskGroup',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('description', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='task',
            name='task_group',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='tasks', to='core.TaskGroup'),
        ),
    ]



from gettext import gettext as _

from logging import getLogger

from rest_framework.serializers import (
    FileField,
    Serializer,
    ValidationError,
)
from pulpcore.plugin.models import Artifact, Repository
from pulpcore.plugin.serializers import (
    DetailRelatedField,
    NoArtifactContentSerializer,
    SingleArtifactContentSerializer,
)


log = getLogger(__name__)


class UploadSerializerFieldsMixin(Serializer):
    """A mixin class that contains fields and methods common to content upload serializers."""

    file = FileField(
        help_text=_("An uploaded file that may be turned into the artifact of the content unit."),
        required=False,
        write_only=True,
    )
    repository = DetailRelatedField(
        help_text=_("A URI of a repository the new content unit should be associated with."),
        required=False,
        write_only=True,
        view_name_pattern=r"repositories(-.*/.*)-detail",
        queryset=Repository.objects.all(),
    )

    def create(self, validated_data):
        """
        Save a GenericContent unit.

        This must be used inside a task that locks on the Artifact and if given, the repository.
        """

        repository = validated_data.pop("repository", None)
        content = super().create(validated_data)

        if repository:
            repository.cast()
            content_to_add = self.Meta.model.objects.filter(pk=content.pk)

            # create new repo version with uploaded package
            with repository.new_version() as new_version:
                new_version.add_content(content_to_add)
        return content

    class Meta:
        fields = ("file", "repository")


class NoArtifactContentUploadSerializer(UploadSerializerFieldsMixin, NoArtifactContentSerializer):
    """A serializer for content types with no Artifact."""

    def create(self, validated_data):
        """Create a new content and remove the already parsed file from validated_data."""
        validated_data.pop("file", None)
        return super().create(validated_data)

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + UploadSerializerFieldsMixin.Meta.fields


class SingleArtifactContentUploadSerializer(
    UploadSerializerFieldsMixin, SingleArtifactContentSerializer
):
    """
    A serializer for content_types with a single Artifact.

    The Artifact can either be specified via it's url, or a new file can be uploaded.
    Additionally a repository can be specified, to which the content unit will be added.

    When using this serializer, the creation of the real content must be wrapped in a task that
    locks the Artifact and when specified, the repository.
    """

    def __init__(self, *args, **kwargs):
        """Initializer for SingleArtifactContentUploadSerializer."""
        super().__init__(*args, **kwargs)
        if self.fields.get("artifact"):
            self.fields["artifact"].required = False

    def validate(self, data):
        """Validate that we have an Artifact or can create one."""

        data = super().validate(data)

        if "file" in data:
            if "artifact" in data:
                raise ValidationError(_("Only one of 'file' and 'artifact' may be specified."))
            data["artifact"] = Artifact.init_and_validate(data.pop("file"))
        elif "artifact" not in data:
            raise ValidationError(_("One of 'file' and 'artifact' must be specified."))

        if "request" not in self.context:
            data = self.deferred_validate(data)

        return data

    def deferred_validate(self, data):
        """
        Validate the content unit by deeply analyzing the specified Artifact.

        This is only called when validating without a request context to prevent stalling
        an ongoing http request.
        It should be overwritten by plugins to extract metadata from the actual content in
        much the same way as `validate`.
        """
        return data

    class Meta(SingleArtifactContentSerializer.Meta):
        fields = (
            SingleArtifactContentSerializer.Meta.fields + UploadSerializerFieldsMixin.Meta.fields
        )



from django.db import transaction

from pulpcore.app.apps import get_plugin_config
from pulpcore.app.models import CreatedResource, PulpTemporaryFile
from pulpcore.app.files import PulpTemporaryUploadedFile
from pulpcore.plugin.models import MasterModel


def general_create_from_temp_file(app_label, serializer_name, temp_file_pk, *args, **kwargs):
    """
    Create a model instance from contents stored in a temporary file.

    A task which executes this function takes the ownership of a temporary file and deletes it
    afterwards. This function calls the function general_create() to create a model instance.
    """
    temp_file = PulpTemporaryFile.objects.get(pk=temp_file_pk)

    data = kwargs.pop("data", {})
    data["file"] = PulpTemporaryUploadedFile.from_file(temp_file.file)

    general_create(app_label, serializer_name, data=data, *args, **kwargs)
    temp_file.delete()


def general_create(app_label, serializer_name, *args, **kwargs):
    """
    Create a model instance.

    Raises:
        ValidationError: If the serializer is not valid

    """
    data = kwargs.pop("data", None)
    context = kwargs.pop("context", {})
    serializer_class = get_plugin_config(app_label).named_serializers[serializer_name]
    serializer = serializer_class(data=data, context=context)
    serializer.is_valid(raise_exception=True)
    serializer.save()
    instance = serializer_class.Meta.model.objects.get(pk=serializer.instance.pk).cast()
    resource = CreatedResource(content_object=instance)
    resource.save()


def general_update(instance_id, app_label, serializer_name, *args, **kwargs):
    """
    Update a model

    The model instance is identified using the app_label, id, and serializer name. The serializer is
    used to perform validation.

    Args:
        id (str): the id of the model
        app_label (str): the Django app label of the plugin that provides the model
        serializer_name (str): name of the serializer class for the model
        data (dict): dictionary whose keys represent the fields of the model and their corresponding
            values.
        partial (bool): When true, only the fields specified in the data dictionary are updated.
            When false, any fields missing from the data dictionary are assumed to be None and
            their values are updated as such.

    Raises:
        :class:`rest_framework.exceptions.ValidationError`: When serializer instance can't be saved
            due to validation error. This theoretically should never occur since validation is
            performed before the task is dispatched.
    """
    data = kwargs.pop("data", None)
    partial = kwargs.pop("partial", False)
    serializer_class = get_plugin_config(app_label).named_serializers[serializer_name]
    instance = serializer_class.Meta.model.objects.get(pk=instance_id).cast()
    serializer = serializer_class(instance, data=data, partial=partial)
    serializer.is_valid(raise_exception=True)
    serializer.save()


def general_delete(instance_id, app_label, serializer_name):
    """
    Delete a model

    The model instance is identified using the app_label, id, and serializer name.

    Args:
        id (str): the id of the model
        app_label (str): the Django app label of the plugin that provides the model
        serializer_name (str): name of the serializer class for the model
    """
    serializer_class = get_plugin_config(app_label).named_serializers[serializer_name]
    instance = serializer_class.Meta.model.objects.get(pk=instance_id).cast()
    instance.delete()


def general_multi_delete(instance_ids):
    """
    Delete a list of model instances in a transaction

    The model instances are identified using the id, app_label, and serializer_name.

    Args:
        instance_ids (list): List of tupels of id, app_label, serializer_name
    """
    instances = []
    for instance_id, app_label, serializer_name in instance_ids:
        serializer_class = get_plugin_config(app_label).named_serializers[serializer_name]
        instance = serializer_class.Meta.model.objects.get(pk=instance_id)
        if isinstance(instance, MasterModel):
            instance = instance.cast()
        instances.append(instance)
    with transaction.atomic():
        for instance in instances:
            instance.delete()



import os
from uuid import uuid4

from django.conf import settings
from django.core.files import locks
from django.core.files.move import file_move_safe
from django.core.files.storage import FileSystemStorage


class FileSystem(FileSystemStorage):
    """
    Django's FileSystemStorage with modified _save() and get_available_name behaviors

    The _save() will check if the file is saved in WORKING_DIRECTORY first. If it is, a move is
    used. This will move all files created by the Downloaders and uploaded files from the user. If
    it is saved in-memory or outside of WORKING_DIRECTORY, the data is written/copied in chunks to
    the new location.
    """

    def get_available_name(self, name, max_length=None):
        """
        Returns a filename for the file even if it already exists.

        Content adressable storage must be saved with the expected filename.

        Args:
            name (string): Requested file name
            max_length (int): Maximum length of the filename. Not used in this implementation.

        Returns:
            Name of the file.
        """
        return name

    def _save(self, name, content, max_length=None):
        """
        Create dirs to the destination, move the file if already in MEDIA_ROOT, or copy otherwise.

        Args:
            name (str): Target path to which the file is copied.
            content (File): Source file object.
            max_length (int): Maximum supported length of file name.

        Returns:
            str: Final storage path.
        """
        full_path = self.path(name)

        # Create any intermediate directories that do not exist.
        directory = os.path.dirname(full_path)
        try:
            if self.directory_permissions_mode is not None:
                # os.makedirs applies the global umask, so we reset it,
                # for consistency with file_permissions_mode behavior.
                old_umask = os.umask(0)
                try:
                    os.makedirs(directory, self.directory_permissions_mode, exist_ok=True)
                finally:
                    os.umask(old_umask)
            else:
                os.makedirs(directory, exist_ok=True)
        except FileExistsError:
            raise FileExistsError("%s exists and is not a directory." % directory)

        try:
            if hasattr(content, "temporary_file_path") and content.temporary_file_path().startswith(
                str(settings.WORKING_DIRECTORY)
            ):
                file_move_safe(content.temporary_file_path(), full_path)
            else:
                # This is a normal uploaded file that we can stream.

                # The current umask value is masked out by os.open!
                fd = os.open(full_path, self.OS_OPEN_FLAGS, 0o666)
                _file = None
                try:
                    locks.lock(fd, locks.LOCK_EX)
                    for chunk in content.chunks():
                        if _file is None:
                            mode = "wb" if isinstance(chunk, bytes) else "wt"
                            _file = os.fdopen(fd, mode)
                        _file.write(chunk)
                finally:
                    locks.unlock(fd)
                    if _file is not None:
                        _file.close()
                    else:
                        os.close(fd)
        except FileExistsError:
            # It's a content addressable store so if the file is already in place we can do nothing
            pass

        if self.file_permissions_mode is not None:
            os.chmod(full_path, self.file_permissions_mode)

        # Store filenames with forward slashes, even on Windows.
        return str(name).replace("\\", "/")


def get_artifact_path(sha256digest):
    """
    Determine the relative path where a file backing the Artifact should be stored.

    Args:
        sha256digest (str): sha256 digest of the file for the Artifact

    Returns:
        A string representing the absolute path where a file backing the Artifact should be
        stored
    """
    return os.path.join("artifact", sha256digest[:2], sha256digest[2:])


def get_temp_file_path(pulp_id):
    """
    Determine the relative path where a file backing the PulpTemporaryFile should be stored.

    Args:
        pulp_id (uuid): An identifier identifying the file for the PulpTemporaryFile

    Returns:
        A string representing the path where a file backing the PulpTemporaryFile should be
        stored. This path is relative to the storage.
    """
    pulp_id_str = str(pulp_id)
    return os.path.join("tmp/files", pulp_id_str[:2], pulp_id_str[2:])


def get_upload_chunk_file_path(pulp_id):
    """
    Determine the relative path where a file backing an uploaded chunk should be stored.

    Args:
        pulp_id (uuid): An identifier identifying the file for UploadChunk
    Returns:
        A string representing the relative path where a file backing UploadChunk should be
        stored
    """
    return os.path.join(settings.CHUNKED_UPLOAD_DIR, str(pulp_id))


def get_tls_path(model, name):
    """
    Determine storage location as: MEDIA_ROOT/tls/<model>/<id>/<name>.

    Args:
        model (pulpcore.app.models.Model): The model object.
        name (str): The (unused) input file path.

    Returns:
        str: An absolute (base) path
    """
    return os.path.join(settings.MEDIA_ROOT, "tls", type(model).__name__, str(uuid4()), name)



from rest_access_policy import AccessPolicy
from django.db.utils import ProgrammingError

from pulpcore.app.models import AccessPolicy as AccessPolicyModel
from pulpcore.app.util import get_view_urlpattern


class AccessPolicyFromDB(AccessPolicy):
    """
    An AccessPolicy that loads statements from an `AccessPolicy` model instance.
    """

    def get_policy_statements(self, request, view):
        """
        Return the policy statements from an AccessPolicy instance matching the viewset name.

        This is an implementation of a method that will be called by
        `rest_access_policy.AccessPolicy`. See the drf-access-policy docs for more info:

        https://rsinger86.github.io/drf-access-policy/loading_external_source/

        The `pulpcore.plugin.models.AccessPolicy` instance is looked up by the `viewset_name`
        attribute using::

            AccessPolicyModel.objects.get(viewset_name=get_view_urlpattern(view))

        If a matching `pulpcore.plugin.models.AccessPolicy` cannot be found, a default behavior of
        allowing only admin users to perform any operation is used. This fallback allows the Pulp
        RBAC implementation to be turned on endpoint-by-endpoint with less effort.

        Args:
            request (rest_framework.request.Request): The request being checked for authorization.
            view (subclass rest_framework.viewsets.GenericViewSet): The view name being requested.

        Returns:
            The access policy statements in drf-access-policy policy structure.
        """
        try:
            viewset_name = get_view_urlpattern(view)
            access_policy_obj = AccessPolicyModel.objects.get(viewset_name=viewset_name)
        except (AccessPolicyModel.DoesNotExist, AttributeError, ProgrammingError):
            default_statement = [{"action": "*", "principal": "admin", "effect": "allow"}]
            policy = getattr(view, "DEFAULT_ACCESS_POLICY", {"statements": default_statement})
            return policy["statements"]
        else:
            return access_policy_obj.statements



import logging
import uuid

from pathlib import Path

import pytest

from pulpcore.client.pulp_file import (
    ContentFilesApi,
    RepositoriesFileApi,
    RepositoriesFileVersionsApi,
    RemotesFileApi,
)
from pulp_smash.pulp3.utils import gen_repo

from pulpcore.tests.functional.api.using_plugin.utils import (
    gen_file_client,
)


_logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def file_client():
    return gen_file_client()


@pytest.fixture(scope="session")
def content_file_api_client(file_client):
    return ContentFilesApi(file_client)


@pytest.fixture(scope="session")
def file_repo_api_client(file_client):
    return RepositoriesFileApi(file_client)


@pytest.fixture(scope="session")
def file_repo_version_api_client(file_client):
    return RepositoriesFileVersionsApi(file_client)


@pytest.fixture
def file_repo(file_repo_api_client, gen_object_with_cleanup):
    return gen_object_with_cleanup(file_repo_api_client, gen_repo())


@pytest.fixture(scope="session")
def file_remote_api_client(file_client):
    return RemotesFileApi(file_client)


@pytest.fixture(scope="session")
def file_fixtures_root():
    return Path(__file__).parent / "fixtures"


@pytest.fixture
def file_fixture_server_ssl_client_cert_req(
    ssl_ctx_req_client_auth, file_fixtures_root, gen_fixture_server
):
    yield gen_fixture_server(file_fixtures_root, ssl_ctx_req_client_auth)


@pytest.fixture
def file_fixture_server_ssl(ssl_ctx, file_fixtures_root, gen_fixture_server):
    yield gen_fixture_server(file_fixtures_root, ssl_ctx)


@pytest.fixture
def file_fixture_server(file_fixtures_root, gen_fixture_server):
    yield gen_fixture_server(file_fixtures_root, None)


@pytest.fixture
def file_fixture_gen_remote(file_fixture_server, file_remote_api_client, gen_object_with_cleanup):
    def _file_fixture_gen_remote(*, fixture_name, policy, **kwargs):
        url = file_fixture_server.make_url(f"/{fixture_name}/PULP_MANIFEST")
        kwargs.update({"url": str(url), "policy": policy, "name": str(uuid.uuid4())})
        return gen_object_with_cleanup(file_remote_api_client, kwargs)

    yield _file_fixture_gen_remote


@pytest.fixture
def file_fixture_gen_remote_ssl(
    file_fixture_server_ssl,
    file_remote_api_client,
    tls_certificate_authority_cert,
    gen_object_with_cleanup,
):
    def _file_fixture_gen_remote_ssl(*, fixture_name, policy, **kwargs):
        url = file_fixture_server_ssl.make_url(f"/{fixture_name}/PULP_MANIFEST")
        kwargs.update(
            {
                "url": str(url),
                "policy": policy,
                "name": str(uuid.uuid4()),
                "ca_cert": tls_certificate_authority_cert,
            }
        )
        return gen_object_with_cleanup(file_remote_api_client, kwargs)

    yield _file_fixture_gen_remote_ssl


@pytest.fixture
def file_fixture_gen_remote_client_cert_req(
    file_fixture_server_ssl_client_cert_req,
    file_remote_api_client,
    tls_certificate_authority_cert,
    client_tls_certificate_cert_pem,
    client_tls_certificate_key_pem,
    gen_object_with_cleanup,
):
    def _file_fixture_gen_remote_client_cert_req(*, fixture_name, policy, **kwargs):
        url = file_fixture_server_ssl_client_cert_req.make_url(f"/{fixture_name}/PULP_MANIFEST")
        kwargs.update(
            {
                "url": str(url),
                "policy": policy,
                "name": str(uuid.uuid4()),
                "ca_cert": tls_certificate_authority_cert,
                "client_cert": client_tls_certificate_cert_pem,
                "client_key": client_tls_certificate_key_pem,
            }
        )
        return gen_object_with_cleanup(file_remote_api_client, kwargs)

    yield _file_fixture_gen_remote_client_cert_req


@pytest.fixture
def file_fixture_gen_file_repo(file_repo_api_client, gen_object_with_cleanup):
    """A factory to generate a File Repository with auto-deletion after the test run."""

    def _file_fixture_gen_file_repo(**kwargs):
        return gen_object_with_cleanup(file_repo_api_client, kwargs)

    yield _file_fixture_gen_file_repo



import tempfile

from unittest.mock import Mock

from django.test import TestCase

from pulpcore.content import Handler
from pulpcore.plugin.models import Artifact, Content, ContentArtifact


class HandlerSaveContentTestCase(TestCase):
    def setUp(self):
        self.c1 = Content.objects.create()
        self.ca1 = ContentArtifact.objects.create(
            artifact=None, content=self.c1, relative_path="c1"
        )
        self.ra1 = Mock(content_artifact=self.ca1)
        self.c2 = Content.objects.create()
        self.ca2 = ContentArtifact.objects.create(
            artifact=None, content=self.c2, relative_path="c2"
        )
        self.ra2 = Mock(content_artifact=self.ca2)

    def download_result_mock(self):
        dr = Mock()
        dr.artifact_attributes = {"size": 0}
        for digest_type in Artifact.DIGEST_FIELDS:
            dr.artifact_attributes[digest_type] = "abc123"
        tmp_file = tempfile.NamedTemporaryFile(delete=False)
        dr.path = tmp_file.name
        return dr

    def test_save_artifact(self):
        """Artifact needs to be created."""
        cch = Handler()
        new_artifact = cch._save_artifact(self.download_result_mock(), self.ra1)
        c1 = Content.objects.get(pk=self.c1.pk)
        self.assertIsNotNone(new_artifact)
        self.assertEqual(c1._artifacts.get().pk, new_artifact.pk)

    def test_save_artifact_artifact_already_exists(self):
        """Artifact turns out to already exist."""
        cch = Handler()
        new_artifact = cch._save_artifact(self.download_result_mock(), self.ra1)

        existing_artifact = cch._save_artifact(self.download_result_mock(), self.ra2)
        c2 = Content.objects.get(pk=self.c2.pk)
        self.assertEqual(existing_artifact.pk, new_artifact.pk)
        self.assertEqual(c2._artifacts.get().pk, existing_artifact.pk)
